# -*- coding: utf-8 -*-
"""
/***************************************************************************
 Topomime
                                 A QGIS plugin
 Topomime Plugin
                              -------------------
        begin                : 2014-12-06
        git sha              : $Format:%H$
        copyright            : (C) 2015 by Éric Mermet (Éhess)
        email                : eric.mermet@ehess.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""


from PyQt4.QtCore import *
from PyQt4.QtGui import *

from qgis.core import *
from qgis.utils import*

# Initialize Qt resources from file resources.py
import resources_rc
# Import the code for the dialog
from topomime_dialog import TopomimeDialog
import os.path

class Topomime:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'Topomime_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)

        # Create the dialog (after translation) and keep reference
        self.dlg = TopomimeDialog()

        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&Topomime')
        # TODO: We are going to let the user set this up in a future iteration
        self.toolbar = self.iface.addToolBar(u'Topomime')
        self.toolbar.setObjectName(u'Topomime')

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('Topomime', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/Topomime/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u'Topomime Plugin'),
            callback=self.run,
            parent=self.iface.mainWindow())


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginMenu(
                self.tr(u'&Topomime'),
                action)
            self.iface.removeToolBarIcon(action)


    def run(self):
        # show the dialog
        self.dlg.show()
        self.dlg.cbLayer.clear()
        self.dlg.setFixedSize(480, 570)

        QObject.connect(self.dlg.cbLayer, SIGNAL("currentIndexChanged(QString)"), self.cbLayerEvent)

        QObject.connect(self.dlg.cbSkeletise, SIGNAL("stateChanged(int)"), self.cbActiveSkeletiseEvent)
        QObject.connect(self.dlg.cbSkeletiseAttribute, SIGNAL("currentIndexChanged(QString)"), self.cbSkeletiseAttributeEvent)

        QObject.connect(self.dlg.cbAttribute2, SIGNAL("stateChanged(int)"), self.cbcbAttribute2Event)
        QObject.connect(self.dlg.cbLevenshteinDistance, SIGNAL("stateChanged(int)"), self.cbLevenshteinDistanceEvent)
        QObject.connect(self.dlg.bChooseFileCorpus, SIGNAL("pressed ()"), self.bChooseFileCorpus)

        QObject.connect(self.dlg.cbSpatialAggregation, SIGNAL("stateChanged(int)"), self.cbSpatialAggregationEvent)
        QObject.connect(self.dlg.bOk, SIGNAL("pressed ()"), self.compute)

        layers = QgsMapLayerRegistry.instance().mapLayers().values()
        for layer in layers:
            if layer.type() == QgsMapLayer.VectorLayer:
                self.dlg.cbLayer.addItem(layer.name(), layer)

        # Run the dialog event loop
        #result = self.dlg.exec_()

        # See if OK was pressed
        #if result == 1:

    def compute(self):

        # compute treatments
        if self.dlg.cbSkeletise.isChecked():
            print "Pre-treatment skeletisation"
            self.dlg.txtLog.setText("Pre-treatment skeletisation")

            indexCbLayer = self.dlg.cbLayer.currentIndex()
            layer = self.dlg.cbLayer.itemData(indexCbLayer)

            indexAttribute = self.dlg.cbSkeletiseAttribute.currentIndex()
            field = self.dlg.cbSkeletiseAttribute.itemData(indexAttribute)

            vowels = self.dlg.txVowels.toPlainText()
            vowels = vowels.split()

            newAttributeSK = self.dlg.txFieldNameSK.toPlainText()

            #Add new field for skeleton attribute
            caps = layer.dataProvider().capabilities()
            if caps & QgsVectorDataProvider.AddAttributes:
                res = layer.dataProvider().addAttributes([QgsField(newAttributeSK, QVariant.String)])

            #Update modified layer
            layer.updateFields()

            #Iterate over features and apply skeletisation
            iter = layer.getFeatures()
            cpt = 0

            # Enter editing mode
            layer.startEditing()

            for feature in iter:
                idx = layer.fieldNameIndex(field.name())
                topo = feature.attributes()[idx]

                #Test if null value
                if topo:

                    topo = topo.lower()

                    idxNewAttribute = layer.fieldNameIndex(newAttributeSK)
                    #print newAttributeSK + str(idx)
                    sk_topo = ''.join(char for char in topo if char.lower() not in vowels)
                    #print sk_topo + str(feature.id()) + str(idxNewAttribute)
                    layer.changeAttributeValue(feature.id(), idxNewAttribute, sk_topo)

                    cpt = cpt + 1
                    if cpt % 10000 == 0:
                        print str(cpt) #+ ' / ' + len(layer.getFeatures())

            # Commit changes
            layer.commitChanges()

        if self.dlg.cbLevenshteinDistance.isChecked():
            print "Compute Levenshtein distance"

            #Get selected layer
            indexCbLayer = self.dlg.cbLayer.currentIndex()
            layer = self.dlg.cbLayer.itemData(indexCbLayer)

            #Get selected attributes
            indexAttribute1 = self.dlg.cbLevenshteinAttribute1.currentIndex()
            field1 = self.dlg.cbLevenshteinAttribute1.itemData(indexAttribute1)

            if self.dlg.cbAttribute2.isChecked() :
                indexAttribute2 = self.dlg.cbLevenshteinAttribute2.currentIndex()
                field2 = self.dlg.cbLevenshteinAttribute2.itemData(indexAttribute2)

            #Get Corpus file
            fileCorpusPath = self.dlg.txFileCorpus.toPlainText()
            #Open file and read lines
            f = open(fileCorpusPath)
            #Pre treatment on corpus, skeletise if needed
            lines = f.readlines()
            corpus_total = []

            #Get vowels if needed
            if self.dlg.cbSkeletiseCorpus.isChecked():
                vowels = self.dlg.txVowels.toPlainText()
                vowels = vowels.split()

            for mot_corpus in lines:
                mot_corpus = mot_corpus.rstrip('\n')        #filter \n
                mot_corpus = mot_corpus.lower()
                mot_corpus = mot_corpus[:len(mot_corpus)-1]       # '-1' -> filter end of line

                if self.dlg.cbSkeletiseCorpus.isChecked():
                    mot_corpus = ''.join(char for char in mot_corpus if char.lower() not in vowels)

                corpus_total.append(mot_corpus)

            print 'Corpus from file :'
            print corpus_total

            #Get levenshtein distance value
            levenshteinDistanceValue = self.dlg.sbValueLevenshtein.value()

            #Create a new layer with the same schema as working layer
            title_new_layer = layer.name() + '_levenshtein_' + str(levenshteinDistanceValue)
            layerLevenshtein = self.createNewLayer(layer, title_new_layer)

            #Get capabilities on new layer
            caps2 = layerLevenshtein.dataProvider().capabilities()

            #Add attribute1 to new layer
            if caps2 & QgsVectorDataProvider.AddAttributes:
                layerLevenshtein.startEditing()
                res = layerLevenshtein.dataProvider().addAttributes([QgsField('mot_corp_1', QVariant.String)])
                res = layerLevenshtein.dataProvider().addAttributes([QgsField('d_levens_1', QVariant.Double)])
                #If attribute 2 is activated
                if self.dlg.cbAttribute2.isChecked():
                    res = layerLevenshtein.dataProvider().addAttributes([QgsField('mot_corp_2', QVariant.String)])
                    res = layerLevenshtein.dataProvider().addAttributes([QgsField('d_levens_2', QVariant.Double)])

                #Update layer with new fields
                layerLevenshtein.updateFields()
                layerLevenshtein.commitChanges()

            #Get features from the selected layer
            iterAllFeaturesLayerSource = layer.getFeatures()

            cpt_feature = 0
            list_features_to_add = []
            for feature in iterAllFeaturesLayerSource:

                new_feat = 'null'

                idx_nom = layer.fieldNameIndex(field1.name())
                stringTopo = feature.attributes()[idx_nom]

                if not stringTopo:
                    stringTopo = ''

                stringTopo = stringTopo.lower()

                if self.dlg.cbAttribute2.isChecked():
                    idx_nom2 = layer.fieldNameIndex(field2.name())
                    stringTopo2 = feature.attributes()[idx_nom2]

                    if not stringTopo:
                        stringTopo2 = ''

                    stringTopo2 = stringTopo.lower()

                cpt_feature = cpt_feature + 1
                if cpt_feature % 5000 == 0:
                    print str(cpt_feature) #+ ' / ' + len(layer.getFeatures())

                flagDetectInAttrib1 = False
                if stringTopo:

                    for mot_corpus in corpus_total:

                        for word in stringTopo.split():

                            valueLevenshtein = self.levenshtein(mot_corpus, word)

                            if valueLevenshtein >= levenshteinDistanceValue:
                                flagDetectInAttrib1 = True
                                print "(attr1) : (" + str(feature.id()) + ") " + word + " // " + mot_corpus + "(c) -> " + str(valueLevenshtein) + " > " + str(levenshteinDistanceValue)
                                #add feature in new layer

                                new_feat = self.copyFeatureAndSetFeatureSchema(feature, layerLevenshtein, layer)

                                new_feat['mot_corp_1'] = mot_corpus
                                new_feat['d_levens_1'] = valueLevenshtein

                        #Compute levenshtein on attribute 2 if checked
                        if self.dlg.cbAttribute2.isChecked():

                            for word2 in stringTopo2.split():

                                valueLevenshtein2 = self.levenshtein(mot_corpus, word2)

                                if valueLevenshtein2 >= levenshteinDistanceValue:
                                    print "(attr2) : (" + str(feature.id()) + ") " + word2 + " // " + mot_corpus + "(c) -> " + str(valueLevenshtein2) + ' > ' + str(levenshteinDistanceValue)

                                    if flagDetectInAttrib1 == False:    #new_feat does not exist
                                        new_feat = self.copyFeatureAndSetFeatureSchema(feature, layerLevenshtein, layer)
                                        new_feat['mot_corp_1'] = 'none'
                                        new_feat['d_levens_1'] = 0.0

                                    new_feat['mot_corp_2'] = mot_corpus
                                    new_feat['d_levens_2'] = valueLevenshtein2

                        #else:
                        #    new_feat['mot_corp_2'] = 'none'
                        #    new_feat['d_levens_2'] = 0.0

                    if new_feat != 'null':
                        list_features_to_add.append(new_feat)

            layerLevenshtein.dataProvider().addFeatures(list_features_to_add)
            layerLevenshtein.commitChanges()

            f.close()

            #add new layer in memory
            QgsMapLayerRegistry.instance().addMapLayer(layerLevenshtein)

            print 'add vector layer ok'

        if self.dlg.cbSpatialAggregation.isChecked():
            print "Spatial Aggregation"

            #Select if layer have to be cascaded with layer levenshtein or choose a new layer if cb levenshtein is not checked
            if self.dlg.cbLevenshteinDistance.isChecked():
                workingLayer = layerLevenshtein
            else:
                #Get selected layer
                indexCbLayer = self.dlg.cbLayer.currentIndex()
                workingLayer = self.dlg.cbLayer.itemData(indexCbLayer)

            print workingLayer.name()

            #Get levenshtein distance value
            spatialDistanceValue = self.dlg.sbSpatialAggreg.value()

            title_layer_aggreg = workingLayer.name() + '_spatialAggreg_' + str(spatialDistanceValue) + 'm'
            layerSpatialAggreg = self.createNewLayer(workingLayer, title_layer_aggreg)

            iterSpatialAggregation1 = workingLayer.getFeatures()

            #Build spatial index
            index = QgsSpatialIndex()
            for featureAggregIndex in iterSpatialAggregation1:
                index.insertFeature(featureAggregIndex)

            #Compute nearest neighbors from spatial index and test distance <= spatialDistanceValue
            iterSpatialAggregation2 = workingLayer.getFeatures()
            #list_features_to_add = []
            list_features_to_add = set()
            for featureAggreg in iterSpatialAggregation2:

                #For 10 nearest neighbors
                nearest = index.nearestNeighbor(featureAggreg.geometry().centroid().asPoint(), 10)
                #print nearest[1]

                for i in range(1, 10):

                    fit2 = workingLayer.getFeatures(QgsFeatureRequest().setFilterFid(nearest[i]))
                    ftr = QgsFeature()
                    fit2.nextFeature(ftr)

                    if featureAggreg.geometry().centroid().distance( ftr.geometry().centroid() ) <= spatialDistanceValue:

                        if self.dlg.cbOnlyIfAttr1Different.isChecked():

                            #Test : Attribute 1 from the two compared features must be different

                            idx_Attr1_featureA = workingLayer.fieldNameIndex('mot_corp_1')
                            attr1_featureA = featureAggreg.attributes()[idx_Attr1_featureA]

                            idx_Attr1_featureB = workingLayer.fieldNameIndex('mot_corp_1')
                            attr1_featureB = ftr.attributes()[idx_Attr1_featureB]

                            if attr1_featureA != attr1_featureB:
                                #Aggregate only if different

                                list_features_to_add.add(featureAggreg.id())
                                list_features_to_add.add(ftr.id())

                        else:
                            list_features_to_add.add(featureAggreg.id())

            #Now add feature
            iterSpatial_trueadd = workingLayer.getFeatures()
            list_features = []
            #list_features_to_add = []
            for this_feat in iterSpatial_trueadd:
                if this_feat.id() in list_features_to_add:
                    list_features.append(this_feat)

            layerSpatialAggreg.dataProvider().addFeatures(list_features)
            layerSpatialAggreg.commitChanges()

            #add new layer in memory
            QgsMapLayerRegistry.instance().addMapLayer(layerSpatialAggreg)

    def copyFeatureAndSetFeatureSchema(self, featureToCopy, layerTarget, layerSource):

        #Build a new feature
        new_feat = QgsFeature()
        new_feat.setGeometry(featureToCopy.geometry())

        new_idx = 0
        fields = layerTarget.pendingFields()
        new_feat.setFields(fields, True)
        for field in layerSource.pendingFields():
            #print str(new_idx) + ' ' + field.name() + ' ' + str(feature.attributes()[new_idx])
            new_feat[field.name()] = featureToCopy.attributes()[new_idx]
            new_idx = new_idx + 1

        return new_feat

    def createNewLayer(self, layerToCopy, title):

        # create layer
        if layerToCopy.type() == QgsMapLayer.VectorLayer:

            #Detect geom type
            if layerToCopy.wkbType() == QGis.WKBPoint:
                type = "Point"

            if layerToCopy.wkbType() == QGis.WKBLineString:
                type = "LineString"

            if layerToCopy.wkbType() == QGis.WKBMultiLineString:
                type = "LineString"

            #Type LineString or MultiLineString BDTopo IGN ???
            if layerToCopy.wkbType() == -2147483646:
                type = "LineString"

            if layerToCopy.wkbType() == QGis.WKBPolygon:
                type = "Polygon"

            if layerToCopy.wkbType() == QGis.WKBMultiPolygon:
                type = "MultiPolygon"

            #Create a new layer and get provider
            vl = QgsVectorLayer(type, title, "memory")
            pr = vl.dataProvider()

            # Enter editing mode
            vl.startEditing()

            # add fields
            fieldsToCopy = layerToCopy.pendingFields()
            tabFields = []
            for field in fieldsToCopy:
                tabFields.append(field)

            pr.addAttributes(tabFields)

            # Commit changes
            #vl.commitChanges()

            return vl

    def cbActiveSkeletiseEvent(self):

        self.dlg.gbPretreatments.setEnabled(self.dlg.cbSkeletise.isChecked())

    def cbLevenshteinDistanceEvent(self):

        self.dlg.gbLevenshteinDistance.setEnabled(self.dlg.cbLevenshteinDistance.isChecked())

    def cbSpatialAggregationEvent(self):

        self.dlg.gbSpatialAggregation.setEnabled(self.dlg.cbSpatialAggregation.isChecked())

    def cbLayerEvent(self):
        #get attributes of selected layer
        self.dlg.cbSkeletiseAttribute.clear()
        self.dlg.cbLevenshteinAttribute1.clear()
        self.dlg.cbLevenshteinAttribute2.clear()

        if self.dlg.cbLayer.count() > 0:

            indexCbLayer = self.dlg.cbLayer.currentIndex()
            layer = self.dlg.cbLayer.itemData(indexCbLayer)

            if layer:

                fields = layer.pendingFields()
                for field in fields:
                    if field.typeName() == 'String':        #add only String type in comboboxes
                        self.dlg.cbSkeletiseAttribute.addItem(field.name(), field)
                        #self.dlg.cbLevenshteinAttribute1.addItem(field.name(), field)
                        #self.dlg.cbLevenshteinAttribute2.addItem(field.name(), field)

                self.dlg.txFieldNameSK.setText("SK_"+fields[0].name())
                #QMessageBox.information(self.iface.mainWindow(),"hello world","%s has %d features." %(layer.name(),layer.featureCount()))

    def cbSkeletiseAttributeEvent(self):
        indexAttribute = self.dlg.cbSkeletiseAttribute.currentIndex()
        if not self.dlg.cbSkeletiseAttribute.count():
            return
        fieldSK = self.dlg.cbSkeletiseAttribute.itemData(indexAttribute)

        self.dlg.cbLevenshteinAttribute1.clear()
        self.dlg.cbLevenshteinAttribute2.clear()

        sk_attribute = 'SK_'+fieldSK.name()
        sk_attribute = sk_attribute[:10]

        self.dlg.txFieldNameSK.setText(sk_attribute)

        indexCbLayer = self.dlg.cbLayer.currentIndex()
        layer = self.dlg.cbLayer.itemData(indexCbLayer)

        fields = layer.pendingFields()
        for field in fields:
            if field.typeName() == 'String':        #add only String type in comboboxes
                self.dlg.cbLevenshteinAttribute1.addItem(field.name(), field)
                self.dlg.cbLevenshteinAttribute2.addItem(field.name(), field)

        self.dlg.cbLevenshteinAttribute1.addItem(sk_attribute, field)
        self.dlg.cbLevenshteinAttribute2.addItem(sk_attribute, field)

    def bChooseFileCorpus(self):
        print 'push button tool'
        #Tk().withdraw()
        #filename = askopenfilename()

        filename = QFileDialog.getOpenFileName()

        self.dlg.txFileCorpus.setText(filename)

    def cbcbAttribute2Event(self):
        self.dlg.cbLevenshteinAttribute2.setEnabled(self.dlg.cbAttribute2.isChecked())

    def levenshtein(self, seq1, seq2):
        oneago = None
        thisrow = range(1, len(seq2) + 1) + [0]
        for x in xrange(len(seq1)):
            twoago, oneago, thisrow = oneago, thisrow, [0] * len(seq2) + [x + 1]
            for y in xrange(len(seq2)):
                delcost = oneago[y] + 1
                addcost = thisrow[y - 1] + 1
                subcost = oneago[y - 1] + (seq1[x] != seq2[y])
                thisrow[y] = min(delcost, addcost, subcost)
        return (1.0-thisrow[len(seq2) - 1]/float(len(seq1)))*100.0